from __future__ import annotations

import logging
import pathlib

from lazylfs import pathutils

_logger = logging.getLogger(__name__)


def link(src: pathlib.Path, dst: pathlib.Path,) -> None:
    if not src.is_dir():
        raise ValueError("Expected src to be a directory")

    src_tails = {
        pathlib.Path(path).relative_to(src)
        for path in src.rglob("*")
        if path.is_file() and not path.is_symlink()
    }

    dst.mkdir(exist_ok=True)

    for tail in sorted(src_tails):
        src_path = src / tail
        dst_path = dst / tail
        dst_path.parent.mkdir(parents=True, exist_ok=True)
        _logger.debug("Linking %s", str(tail))

        if not pathutils.ensure_lnk(dst_path, src_path):
            _logger.debug("Path exists and is equivalent, skipping")



from __future__ import annotations

import hashlib
import logging
import os
import pathlib
from typing import (
    Dict,
    Set,
)

_logger = logging.getLogger(__name__)

_INDEX_NAME = ".shasum"


class NotOkError(Exception):
    pass


def _sha256(path: pathlib.Path) -> str:
    h = hashlib.sha256()
    b = bytearray(128 * 1024)
    mv = memoryview(b)
    with path.resolve().open("rb", buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):  # type: ignore
            h.update(mv[:n])
    return h.hexdigest()


def _fingerprint_from_content(path):
    return _sha256(path)


def _fingerprint_from_location(path):
    index = _read_index(path.parent / _INDEX_NAME)
    return index.get(path.name)


def _should_be_indexed(path: pathlib.Path) -> bool:
    return path.is_symlink() and path.is_file() and os.path.isabs(os.readlink(path))


def _read_index(path: pathlib.Path) -> Dict[str, str]:
    if not path.exists():
        return {}
    split_lines = [line.split() for line in path.read_text().splitlines()]
    result = {line[-1]: line[0] for line in split_lines}
    if len(result) != len(split_lines):
        raise RuntimeError("Index contains duplicate entries")
    return result


def _append_to_index(link_path: pathlib.Path) -> None:
    index_path = link_path.parent / _INDEX_NAME
    index = _read_index(index_path)

    if link_path.name in index:
        if index[link_path.name] == _fingerprint_from_content(link_path):
            return
        else:
            raise TypeError("Cannot reassign existing key")

    with index_path.open("a") as f:
        f.write(f"{_fingerprint_from_content(link_path)}  {link_path.name}\n")


def track(paths: Set[pathlib.Path]) -> None:
    for path in paths:
        if _should_be_indexed(path):
            _append_to_index(path)


def _check_index(path):
    index = _read_index(path)

    indexed_names = set(index)
    existing_names = set(
        path.name for path in path.parent.iterdir() if _should_be_indexed(path)
    )

    if indexed_names != existing_names:
        return False

    for name, key_from_location in index.items():
        if _fingerprint_from_content(path.parent / name) != key_from_location:
            return False

    return True


def check(paths: Set[pathlib.Path]) -> None:
    ok = True
    for path in paths:
        if path.name == _INDEX_NAME:
            if _check_index(path):
                continue
        elif _should_be_indexed(path):
            if _fingerprint_from_location(path) == _fingerprint_from_content(path):
                continue
        else:
            if _fingerprint_from_location(path) is None:
                continue

        ok &= False
        _logger.debug("NOK %s", path)

    if not ok:
        raise NotOkError



import itertools
import os
import pathlib
import stat
from typing import Iterator, Union


def _resolve_symlink(src: pathlib.Path) -> pathlib.Path:
    """Resolve the immediate target of a symlink

    Contrast this with :py:meth:`pathlib.Path.resolve` that resolves the final target
    of a symlink, possibly resolving many immediate targets along the way, and with
    :py:func:`os.path.normpath` that will not properly resolve parents when the path
    goes through a symlink to a directory.
    """
    tgt = os.readlink(src)
    if os.path.isabs(tgt):
        return pathlib.Path(tgt)

    head, tail = os.path.split(tgt)
    if tail == "..":
        return (src.parent / tgt).resolve()
    else:
        return (src.parent / head).resolve() / tail


def trace_symlink(path: os.PathLike) -> Iterator[pathlib.Path]:
    """Follow symlink to its final target and yield all hops along the way

    The final target is yielded last, this is the only element yielded that is not
    guaranteed to be a symlink.

    The given symlink is not included in the result (subject to change).

    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     pathlib.Path(tmp, "b").touch()
    ...     pathlib.Path(tmp, "lb").symlink_to("./b")
    ...     pathlib.Path(tmp, "llb").symlink_to("./lb")
    ...     [hop.name for hop in trace_symlink(pathlib.Path(tmp, "llb"))]
    ['lb', 'b']
    """
    path = pathlib.Path(path)
    visited = {path}
    while path.is_symlink():
        path = _resolve_symlink(path)
        if path in visited:
            return
        visited.add(path)
        yield path


def ensure_dir(path: os.PathLike, root: os.PathLike) -> bool:
    """Ensure that the specified directory exists

    Avoids creating paths in unintended locations by refusing to create directories
    outside of :param root:.

    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     path = pathlib.Path(tmp, "a", "e")
    ...     assert ensure_dir(path, path.parents[1]) is True
    ...     assert ensure_dir(path, path.parents[1]) is False

    :param path: location of directory
    :param root: an existing ancestor of the directory
    :return: ``True`` if path was created, ``False`` otherwise
    :raises FileExistsError: if the path exists and is different than what would be
        created by this function.
    """
    path = pathlib.Path(path)
    root = pathlib.Path(root)
    root.lstat()  # Raise early if root does not exist
    path.relative_to(root)  # Raise if not ancestor/descendant

    try:
        st_mode = os.lstat(path).st_mode
    except FileNotFoundError:
        parents = itertools.takewhile(lambda p: p != root, path.parents)
        for parent in reversed(list(parents)):
            parent.mkdir(exist_ok=True)
        path.mkdir(exist_ok=True)
        return True

    if not stat.S_ISDIR(st_mode):
        raise FileExistsError(f"File exists (wrong type): {path}")

    return False


def ensure_lnk(path: os.PathLike, tgt: os.PathLike) -> bool:
    """Ensure that the specified symlink exists

    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     path = pathlib.Path(tmp, "b")
    ...     assert ensure_lnk(path, "./foo") is True
    ...     assert ensure_lnk(path, "./foo") is False

    :param path: location of symlink
    :param tgt: desired target of the symlink
    :return: ``True`` if path was created, ``False`` otherwise
    :raises FileExistsError: if the path exists and is different than what would be
        created by this function.
    """
    try:
        st_mode = os.lstat(path).st_mode
    except FileNotFoundError:
        os.symlink(tgt, path)
        return True

    if not stat.S_ISLNK(st_mode):
        raise FileExistsError(f"File exists (wrong type): {path}")

    if os.readlink(path) != os.fspath(tgt):
        raise FileExistsError(f"File exists (wrong content): {path}")

    return False


def ensure_reg(path: os.PathLike, content: Union[bytes, str]) -> bool:
    """Ensure that the specified file exists

    >>> import tempfile
    >>> with tempfile.TemporaryDirectory() as tmp:
    ...     path = pathlib.Path(tmp, "b")
    ...     assert ensure_reg(path, "Bravo") is True
    ...     assert ensure_reg(path, "Bravo") is False

    :param path: location of file
    :param content: desired content of given file
    :return: ``True`` if path was created, ``False`` otherwise
    :raises FileExistsError: if the path exists and is different than what would be
        created by this function.
    """
    if isinstance(content, str):
        content = content.encode()

    try:
        st_mode = os.lstat(path).st_mode
    except FileNotFoundError:
        with open(path, "xb") as f:
            f.write(content)
        return True

    if not stat.S_ISREG(st_mode):
        raise FileExistsError(f"File exists (wrong type): {path}")

    with open(path, "rb") as f:
        if f.read() != content:
            raise FileExistsError(f"File exists (wrong content): {path}")

    return False



from __future__ import annotations

import logging
import os
import pathlib
import sys
from typing import (
    Union,
    TYPE_CHECKING,
    Set,
    Tuple,
    Iterator,
)

from lazylfs import content, location

_logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    PathT = Union[str, os.PathLike[str], pathlib.Path]


def _collect_paths(includes: Tuple[str, ...]) -> Set[pathlib.Path]:
    if not includes:
        includes = tuple([line.rstrip() for line in sys.stdin.readlines()])

    included: Set[pathlib.Path] = set()
    for top in includes:
        included.update(_find(pathlib.Path(top)))
    return included


def _find(top: pathlib.Path) -> Iterator[pathlib.Path]:
    yield top
    yield from top.rglob("*")


def link(src: PathT, dst: PathT,) -> None:
    """Create links in `dst` to the corresponding files in `src`

    :param src: Directory under which to look for files
    :param dst: Directory under which to create symlinks
    """
    src = pathlib.Path(src).resolve()
    dst = pathlib.Path(dst).resolve()
    location.link(src, dst)


def track(*includes: str,) -> None:
    """Track the checksum of files in the index"""
    content.track(_collect_paths(includes))


NotOkError = content.NotOkError


def check(*includes: str,) -> None:
    """Check the checksum of files against the index

    Exit with non-zero status if a difference is detected or a file could not be
    checked.
    """
    content.check(_collect_paths(includes))


def main():
    import fire  # type: ignore

    logging.basicConfig(level=getattr(logging, os.environ.get("LEVEL", "WARNING")))
    fire.Fire({func.__name__: func for func in [link, track, check]})



#!/usr/bin/env python
import pathlib
from typing import List

import setuptools


def _read(rel_path: str) -> str:
    abs_path = pathlib.Path(__file__).parent / rel_path
    return abs_path.read_text()


def _read_tagline() -> str:
    lines = _read("README.md").splitlines(keepends=False)
    for line in lines:
        if line and line[0] == line[-1] == "*":
            return line[1:-1]


def _read_requirements(name: str) -> List[str]:
    return list(_read(f"requirements/{name}.txt").splitlines())


setuptools.setup(
    name="lazylfs",
    author="AP Ljungquist",
    author_email="ap@ljungquist.eu",
    description=_read_tagline(),
    long_description=_read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/apljungquist/lazylfs",
    packages=setuptools.find_packages("src"),
    package_data={"lazylfs": ["py.typed"]},
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
    ],
    install_requires=_read_requirements("install_requires"),
    extras_require={"cli": _read_requirements("extras_require-cli")},
    package_dir={"": "src"},
    entry_points={"console_scripts": ["lazylfs = lazylfs.cli:main [cli]"]},
)
